/*==============================================================================


   Purpose:    event descriptors


   Author:     Rudolf Berrendorf
               University of Applied Sciences Bonn-Rhein-Sieg
               Computer Science Department
               53754 Sankt Augustin
               rudolf.berrendorf@fh-bonn-rhein-sieg.de


   last mod:   20.06.2000


   DISCLAIMER

                  PCL - The Performance Counter Library


   Copyright (C) 2000
   Forschungszentrum Juelich GmbH, Federal Republic of Germany.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

   Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

     - Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.

     - Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

     - Any publications that result from the use of this software shall
       reasonably refer to the Research Centre's development.

     - All advertising materials mentioning features or use of this software
       must display the following acknowledgement:

           This product includes software developed by Forschungszentrum
           Juelich GmbH, Federal Republic of Germany.

     - Forschungszentrum Juelich GmbH is not obligated to provide the user with
       any support, consulting, training or assistance of any kind with regard
       to the use, operation and performance of this software or to provide
       the user with any updates, revisions or new versions.


   THIS SOFTWARE IS PROVIDED BY FORSCHUNGSZENTRUM JUELICH GMBH "AS IS" AND ANY
   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
   DISCLAIMED. IN NO EVENT SHALL FORSCHUNGSZENTRUM JUELICH GMBH BE LIABLE FOR
   ANY SPECIAL, DIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
   RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
   CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE ACCESS, USE OR PERFORMANCE OF THIS SOFTWARE.
  
==============================================================================*/

#if !defined(PCL_EVENT_DESCR_INCLUDED)
#define PCL_EVENT_DESCR_INCLUDED


/* marker in table if this event is not available/available */
#define PCL_UNAVAIL      0
#define PCL_AVAIL        1


/* For each event, there exists a table entry with the following components:.
   If the event is in some way available on the computer, "available" is
   set to 1.
   If the event is mapped directly to a low-level counter, "composed" is
   set to NOT_COMP and "data.command" contains the appropriate command word
   for the driver. Each entry not equal to -1 is a possible command word. As
   one event may be counted on different low-level counters, several command[i]
   may be different to -1.
   If the event is measured by a combination of some low-level counters,
   "composed" is set to COMPOSED and the rest of the data is stored in
   "composed_data". "composed_data.command" contains the command words and
   "composed.data.operation contains the reduction operation of the
   appropriate low-level counter values (e.g. (+1) * ctr_0 + (-1) * ctr_1).
   Example for non-composed entry:
   {PCL_AVAIL, NOT_COMP, {0x02, 0x03}}
   The event may be counted on counter 0 or counter 1, the command words for
   the low-level driver are 0x02 (for counter 0) and 0x03 (counter 1),
   respectively.
   Example for composed entry:
   {PCL_AVAIL, COMPOSED, {{0x0e, 0x0f}, {1, -1}}}
   The event is composed out of two events. The command words for
   the low-level driver are 0x0e (for counter 0) and 0x0f (counter 1),
   respectively. The resulting counter values (ctr_0, ctr_1) are connected in
   the form: (+1) * ctr_0 + (-1) * ctr_1
 */

/* additional bits to change operand type */
#define OP_FP_1         128u	       /* change int operand 1 to floating
				          point */
#define OP_FP_2         256u	       /* change int operand 2 to floating
				          point */
#define OP_FP_MASK      0x3f	       /* mask out OP_FP_x */
#define OP_FP_TEST_MASK 0x180	       /* mask to test if FP bit is set */

/* operations (might be combined with OP_FP_x)
   operations are strictly applied left to right, i.e. all operations on
   counter 0 (if used) followed by all operations on counter 1 (if used) etc.
*/
#define OP_ADD         1u	       /* int add value */
#define OP_SUB         2u	       /* int subtract value */
#define OP_FP_STORE    3u	       /* floating point store */
#define OP_FP_ADD      4u	       /* floating point add */
#define OP_FP_MUL      5u	       /* floating point multiply */
#define OP_FP_DIV      6u	       /* floating point divide */
#define OP_FP_DIV_MHZ  7u	       /* floating point divide by value and
				          multiply with MHz */
#define OP_FP_REZ      8u	       /* add 1.0/value */
#define OP_FP_REZ_MHZ  9u	       /* add MHz/value */
#define OP_NOOP        10u             /* no operation */

#define OP_NO         (~0u)	       /* no entry */

#define COMPOSED 1		       /* event composed */
#define NOT_COMP 0		       /* event not composed */


typedef struct
{
  unsigned int available:1;	       /* event is available in HW/driver? */
  unsigned int composed:1;	       /* event is composed? */
  union
  {
    struct
    {
      PCL_DRIVER_COMMAND_TYPE command[PCL_MAX_COUNTER];	/* command values */
      unsigned int operation[PCL_MAX_COUNTER];	/* combine oper. */
    }
    composed_data;
    PCL_DRIVER_COMMAND_TYPE command[PCL_MAX_COUNTER];	/* command value */
  }
  data;
}
PCL_event_descr_t;

#endif				       /* !defined(PCL_EVENT_DESCR_INCLUDED) */


/*============================================================================*
 *                             that's all folks                               *
 *============================================================================*/
