/*==============================================================================


   Purpose:    Performance Counter Library (PCL)


   Author:     Rudolf Berrendorf
               University of Applied Sciences Bonn-Rhein-Sieg
               Computer Science Department
               53754 Sankt Augustin
               rudolf.berrendorf@fh-bonn-rhein-sieg.de


   last mod:   02.07.2002


   DISCLAIMER

                  PCL - The Performance Counter Library


   Copyright (C) 2002
   Rudolf Berrendorf

   Copyright (C) 1999, 2000
   Forschungszentrum Juelich GmbH, Federal Republic of Germany.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

   Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

     - Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.

     - Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

     - Any publications that result from the use of this software shall
       reasonably refer to the Research Centre's development.

     - All advertising materials mentioning features or use of this software
       must display the following acknowledgement:

           This product includes software developed by Forschungszentrum
           Juelich GmbH, Federal Republic of Germany.

     - Forschungszentrum Juelich GmbH is not obligated to provide the user with
       any support, consulting, training or assistance of any kind with regard
       to the use, operation and performance of this software or to provide
       the user with any updates, revisions or new versions.


   THIS SOFTWARE IS PROVIDED BY FORSCHUNGSZENTRUM JUELICH GMBH "AS IS" AND ANY
   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
   DISCLAIMED. IN NO EVENT SHALL FORSCHUNGSZENTRUM JUELICH GMBH BE LIABLE FOR
   ANY SPECIAL, DIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
   RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
   CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE ACCESS, USE OR PERFORMANCE OF THIS SOFTWARE.
  
==============================================================================*/


#include "defines.h"
#include "pcl.h"
#include "determine_mhz.h"
#include "processor_number.h"
#include "state_descr.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#if defined(PCL_MULTITHREADED)
#include <pthread.h>
#endif

#if defined(JAVA_INTERFACE)
/* the include file PCL.h gets automatically generated by a call to javah */
#include "PCL.h"
#include <jni.h>
#endif


/*                                                                            */
/*============================================================================*/
/* Global definitions. Read carefully the description and decide if you want  */
/* to change any of them.                                                     */
/* Defaults are set in such way that you always get correct results on all    */
/* systems. I you are not sure, don't touch the definitions.                  */


/* Our performance counter result values are 64 bit integers. On some systems,
   hardware counters are only 32 bit counters (see below). Therefore, on such
   systems we activate a timer which interrupts and we gather temporary
   results. For such systems, COUNTER_SIZE_32 must be defined (below).
   The timer which gets activated interrupts every PCL_INTERRUPT_INTERVALL
   milliseconds to gather performance data. If you know the MHZ rate or cycle
   time of a processor, this could be more accurately set for every processor.
   Change the following definition if your processor can generate more than 2^32
   events per second (e.g. if your processor is running at 5 Gigaherz:-) */
#define PCL_INTERRUPT_INTERVALL   500     /* interrupt intervall (in msec) */

/* Normally, we decrement a CPU time counter to interrupt after
   PCL_INTERRUPT_INTERVALL msec. Doing so, the signal SIGVTALRM is used.
   If for some special reason you don't like that and prefer elapsed time
   processor time delivering the signal SIGALRM change the following
   line to "#if 0".  Please be aware that for an elapsed cycle counter
   counting is going on even your process is not scheduled by the OS.
   But please be aware, that there are problems with SIGALRM on many systems
   if you do for example I/O.
  */
#if 1
#define ITIMER_MODE   ITIMER_VIRTUAL  /* decrement virtual process time */
#define ITIMER_SIGNAL SIGVTALRM       /* SIGVTALRM is delivered */
#else
/* there are problems with SIGALRM in combination with I/O on several systems */
#define ITIMER_MODE   ITIMER_REAL     /* decrement in real time */
#define ITIMER_SIGNAL SIGALRM         /* SIGALRM is delivered */
#endif


/* Undefine this macro if you have 32-bit counters but your measurements
   don't exceed 32-bit counters. If you undefine this macro, timers don't
   need to be set up and you can save a lot of overhead in PCLstart/PCLstop.
   We undefine this if we are on a SUN and have Java enabled as using SIGALRM 
   conflicts with I/O and using SIGVTALARM conflicts with Java :-(
   Also on Cray systems the generated interrupts every 2 seconds can generate
   for some applications a HUGE overhead (believe me) for reasons we
   don't know. Therefore we switch it off.
*/
#if !((defined(SPARC) && defined(JAVA_INTERFACE)) || defined(ALPHA_CRAY))
#define COUNTER32_ARE_NOT_ENOUGH
#endif


/* Undefine this macro if you don't want to check consistency of function
   arguments, i.e. you are sure always to deliver correct arguments to the
   library interface functions.
*/
#define CHECK_ARGUMENTS

/* a magic value */
#define MAGIC  0xdeadbeef

#include "pcl_hw_driver.h"
#include "event_descr.h"

#if defined(COUNTER_SIZE_32)
#include <signal.h>
#endif


/*                                                                            */
/*============================================================================*/
/* system independent read-only variables                                     */


/* names of events (gets initialized in init_eventnames) */
static const char *(PCL_eventname[PCL_MAX_EVENT]);

/*                                                                            */
/*============================================================================*/
/* System independent variables.
   We put them all in one structure to simplify support for multiple threads.
   Every thread gets its own set of variables.
*/


#if defined(PCL_MULTITHREADED)
/* This is the key to access thread specific descriptor data */
static pthread_key_t descr_key;
#else
/* define one variable of that structure for non-reentrant calls. */
static PCL_state_descr *state_descr;
#endif


/*                                                                            */
/*============================================================================*/
/* forward declarations                                                       */


static int
hw_counter_start(PCL_state_descr *descr);


static int
hw_counter_read(
		PCL_state_descr *descr,
		PCL_CNT_TYPE counter_values[PCL_MAX_COUNTER] /* O: counter values */
		);

static int
hw_counter_stop(PCL_state_descr *descr);

static int
hw_counter_read_and_stop(
	       PCL_state_descr *descr,
	       PCL_CNT_TYPE counter_values[PCL_MAX_COUNTER] /* O: counter values */
	       );


/*                                                                            */
/*============================================================================*/
/* initialize event name array        .                                       */
/* This is done at runtime as we otherwise get errors with dynamic libraries  */


static void
init_eventnames(void)
{
  PCL_eventname[PCL_L1CACHE_READ]        = "PCL_L1CACHE_READ";
  PCL_eventname[PCL_L1CACHE_WRITE]       = "PCL_L1CACHE_WRITE";
  PCL_eventname[PCL_L1CACHE_READWRITE]   = "PCL_L1CACHE_READWRITE";
  PCL_eventname[PCL_L1CACHE_HIT]         = "PCL_L1CACHE_HIT";
  PCL_eventname[PCL_L1CACHE_MISS]        = "PCL_L1CACHE_MISS";
  PCL_eventname[PCL_L1DCACHE_READ]       = "PCL_L1DCACHE_READ";
  PCL_eventname[PCL_L1DCACHE_WRITE]      = "PCL_L1DCACHE_WRITE";
  PCL_eventname[PCL_L1DCACHE_READWRITE]  = "PCL_L1DCACHE_READWRITE";
  PCL_eventname[PCL_L1DCACHE_HIT]        = "PCL_L1DCACHE_HIT";
  PCL_eventname[PCL_L1DCACHE_MISS]       = "PCL_L1DCACHE_MISS";
  PCL_eventname[PCL_L1ICACHE_READ]       = "PCL_L1ICACHE_READ";
  PCL_eventname[PCL_L1ICACHE_WRITE]      = "PCL_L1ICACHE_WRITE";
  PCL_eventname[PCL_L1ICACHE_READWRITE]  = "PCL_L1ICACHE_READWRITE";
  PCL_eventname[PCL_L1ICACHE_HIT]        = "PCL_L1ICACHE_HIT";
  PCL_eventname[PCL_L1ICACHE_MISS]       = "PCL_L1ICACHE_MISS";
  PCL_eventname[PCL_L2CACHE_READ]        = "PCL_L2CACHE_READ";
  PCL_eventname[PCL_L2CACHE_WRITE]       = "PCL_L2CACHE_WRITE";
  PCL_eventname[PCL_L2CACHE_READWRITE]   = "PCL_L2CACHE_READWRITE";
  PCL_eventname[PCL_L2CACHE_HIT]         = "PCL_L2CACHE_HIT";
  PCL_eventname[PCL_L2CACHE_MISS]        = "PCL_L2CACHE_MISS";
  PCL_eventname[PCL_L2DCACHE_READ]       = "PCL_L2DCACHE_READ";
  PCL_eventname[PCL_L2DCACHE_WRITE]      = "PCL_L2DCACHE_WRITE";
  PCL_eventname[PCL_L2DCACHE_READWRITE]  = "PCL_L2DCACHE_READWRITE";
  PCL_eventname[PCL_L2DCACHE_HIT]        = "PCL_L2DCACHE_HIT";
  PCL_eventname[PCL_L2DCACHE_MISS]       = "PCL_L2DCACHE_MISS";
  PCL_eventname[PCL_L2ICACHE_READ]       = "PCL_L2ICACHE_READ";
  PCL_eventname[PCL_L2ICACHE_WRITE]      = "PCL_L2ICACHE_WRITE";
  PCL_eventname[PCL_L2ICACHE_READWRITE]  = "PCL_L2ICACHE_READWRITE";
  PCL_eventname[PCL_L2ICACHE_HIT]        = "PCL_L2ICACHE_HIT";
  PCL_eventname[PCL_L2ICACHE_MISS]       = "PCL_L2ICACHE_MISS";
  PCL_eventname[PCL_TLB_HIT]             = "PCL_TLB_HIT";
  PCL_eventname[PCL_TLB_MISS]            = "PCL_TLB_MISS";
  PCL_eventname[PCL_ITLB_HIT]            = "PCL_ITLB_HIT";
  PCL_eventname[PCL_ITLB_MISS]           = "PCL_ITLB_MISS";
  PCL_eventname[PCL_DTLB_HIT]            = "PCL_DTLB_HIT";
  PCL_eventname[PCL_DTLB_MISS]           = "PCL_DTLB_MISS";
  PCL_eventname[PCL_CYCLES]              = "PCL_CYCLES";
  PCL_eventname[PCL_ELAPSED_CYCLES]      = "PCL_ELAPSED_CYCLES";
  PCL_eventname[PCL_INTEGER_INSTR]       = "PCL_INTEGER_INSTR";
  PCL_eventname[PCL_FP_INSTR]            = "PCL_FP_INSTR";
  PCL_eventname[PCL_LOAD_INSTR]          = "PCL_LOAD_INSTR";
  PCL_eventname[PCL_STORE_INSTR]         = "PCL_STORE_INSTR";
  PCL_eventname[PCL_LOADSTORE_INSTR]     = "PCL_LOADSTORE_INSTR";
  PCL_eventname[PCL_INSTR]               = "PCL_INSTR";
  PCL_eventname[PCL_JUMP_SUCCESS]        = "PCL_JUMP_SUCCESS";
  PCL_eventname[PCL_JUMP_UNSUCCESS]      = "PCL_JUMP_UNSUCCESS";
  PCL_eventname[PCL_JUMP]                = "PCL_JUMP";
  PCL_eventname[PCL_ATOMIC_SUCCESS]      = "PCL_ATOMIC_SUCCESS";
  PCL_eventname[PCL_ATOMIC_UNSUCCESS]    = "PCL_ATOMIC_UNSUCCESS";
  PCL_eventname[PCL_ATOMIC]              = "PCL_ATOMIC";
  PCL_eventname[PCL_STALL_INTEGER]       = "PCL_STALL_INTEGER";
  PCL_eventname[PCL_STALL_FP]            = "PCL_STALL_FP";
  PCL_eventname[PCL_STALL_JUMP]          = "PCL_STALL_JUMP";
  PCL_eventname[PCL_STALL_LOAD]          = "PCL_STALL_LOAD";
  PCL_eventname[PCL_STALL_STORE]         = "PCL_STALL_STORE";
  PCL_eventname[PCL_STALL]               = "PCL_STALL";
  PCL_eventname[PCL_MFLOPS]              = "PCL_MFLOPS";
  PCL_eventname[PCL_IPC]                 = "PCL_IPC";
  PCL_eventname[PCL_L1DCACHE_MISSRATE]   = "PCL_L1DCACHE_MISSRATE";
  PCL_eventname[PCL_L2DCACHE_MISSRATE]   = "PCL_L2DCACHE_MISSRATE";
  PCL_eventname[PCL_MEM_FP_RATIO]        = "PCL_MEM_FP_RATIO";
}


/*                                                                            */
/*============================================================================*/
/* interrupt routine to avoid overflow.                                       */
/* 32-bit performance counters only                                           */


#if defined(COUNTER_SIZE_32) && defined(COUNTER32_ARE_NOT_ENOUGH)

static void
interrupt_handling(
		   int sig
)

{
  int i, nlevel;
  PCL_CNT_TYPE counter_values[PCL_MAX_EVENT_PER_CALL];
  PCL_state_descr *descr;


  /* get state descriptor */
#if defined(PCL_MULTITHREADED)
  descr = pthread_getspecific(descr_key);
#else
  descr = state_descr;
#endif

  /* stop counting and read values */
  if (hw_counter_read_and_stop(descr, counter_values) != PCL_SUCCESS)
    fprintf(stderr, "PCL: problem in interrupt routine (1)\n");

  /* add partial results to current nesting level */
  DEFINE_USE(descr->virtual_counters_used_mask, descr->nesting_level);
  for (i = 0; i < PCL_MAX_COUNTER; ++i)
    descr->virtual_counters[descr->nesting_level][i] += counter_values[i];

  /* restart counting and read out current values as base values
     for all nesting level */
  if ((hw_counter_start(descr) != PCL_SUCCESS)
      || (hw_counter_read(descr, descr->start_values[descr->nesting_level])
	  != PCL_SUCCESS))
    fprintf(stderr, "PCL: problem in interrupt routine (2)\n");

  /* all other nesting levels are reset to the new starting values.
     The final results for each level get updated by the upper level values.
   */
  for(nlevel = descr->nesting_level - 1; nlevel >= 0; --nlevel)
    for(i = 0; i < PCL_MAX_COUNTER; ++i)
      descr->start_values[nlevel][i] = descr->start_values[descr->nesting_level][i];
}

#endif



/*                                                                            */
/*============================================================================*/
/* Start hardware couters.                                                    */
/* This includes setting up the intervall timer and the appropriate signal    */
/* handler to catch timer interrupts.                                         */


static int
hw_counter_start(
		 PCL_state_descr *descr
)

{
  PCL_DRIVER_COMMAND_TYPE commands[PCL_MAX_COUNTER];

  /* 32-bit performance counters */
#if defined(COUNTER_SIZE_32) && defined(COUNTER32_ARE_NOT_ENOUGH)
  static const struct itimerval itimer =
  {{PCL_INTERRUPT_INTERVALL / 1000, 0},
   {PCL_INTERRUPT_INTERVALL / 1000, 0}};
  

  /* set up signal handler */
  if(descr->old_signal_handler != interrupt_handling)
    {
      descr->old_signal_handler = signal(ITIMER_SIGNAL, interrupt_handling);
      if (descr->old_signal_handler == SIG_ERR)
	return PCL_FAILURE;
    }

  /* set up timer to deliver signal */
  if(setitimer(ITIMER_MODE, (struct itimerval *) &itimer, NULL) != 0)
    return PCL_FAILURE;
#endif

  /*--------------------------------------------------------------------------*/
  /* system specific actions */

#if defined(ALPHA_CRAY)

  {
    commands[0] = descr->results[0].command;
    commands[1] = descr->results[1].command;
    commands[2] = descr->results[2].command;
    commands[3] = 0;
  }


  /*--------------------------------------------------------------------------*/

#elif defined(ALPHA_DEC)

  {
    commands[0] = descr->results[0].command;
    commands[1] = descr->results[1].command;
    commands[2] = descr->results[2].command;
    commands[3] = 0;
  }


  /*--------------------------------------------------------------------------*/

#elif defined(MIPS)

  {
    commands[0] = descr->results[0].command;
    commands[1] = descr->results[1].command;
  }


  /*--------------------------------------------------------------------------*/

#elif defined(SPARC)

  {
    commands[0] = descr->results[0].command;
    commands[1] = descr->results[1].command;
    commands[2] = 0;
  }


  /*--------------------------------------------------------------------------*/

#elif defined(POWERPC)

  {
    commands[0] = descr->results[0].command;
    commands[1] = descr->results[1].command;
    commands[2] = descr->results[2].command;
    commands[3] = descr->results[3].command;
    commands[4] = descr->results[4].command;
    commands[5] = descr->results[5].command;
    commands[6] = descr->results[6].command;
    commands[7] = descr->results[7].command;
    commands[8] = descr->results[8].command;
  }


  /*--------------------------------------------------------------------------*/

#elif defined(PENTIUM_LINUX)

  {
    int i;
    
    for(i = 0; i < 19; i++)
      commands[i] = descr->results[i].command;
  }

#endif

  /*--------------------------------------------------------------------------*/

  return PCL_driver_start(descr, PCL_COUNTER_MAX, descr->counter_used_mask,
			  commands, descr->count_mode);
}


/*                                                                            */
/*============================================================================*/
/* read out counters.                                                         */


static int
hw_counter_read(
		PCL_state_descr *descr,
		PCL_CNT_TYPE counter_values[PCL_MAX_COUNTER] /* O: counter values */
)

{
  return PCL_driver_read(descr, PCL_COUNTER_MAX, descr->counter_used_mask, 
			 counter_values);
}


/*                                                                            */
/*============================================================================*/
/* stop counters.                                                             */


static int
hw_counter_stop(PCL_state_descr *descr)

{
  return PCL_driver_stop(descr);
}


/*                                                                            */
/*============================================================================*/
/* Reads out hardware counters and stops counting.                            */


static int
hw_counter_read_and_stop(
	       PCL_state_descr *descr,
	       PCL_CNT_TYPE counter_values[PCL_MAX_COUNTER] /* O: counter values */
)

{
  int res;


  /* read out the current values */
  if((res = hw_counter_read(descr, counter_values)) != PCL_SUCCESS)
    return res;

  /* stop counters */
  return hw_counter_stop(descr);
}


/*                                                                            */
/*============================================================================*/
/* Try to match the event requests with the available counters.               */
/* First, try to match composed events as these are fixed to certain counters.*/
/* Then, try to match integral events to left counters.                       */
/* Not all possible mappings are successful as we have a simple strategy      */
/* might return "not possible" even if there is some combination.             */
/* Should be done better in the future.                                       */


static int
map_low_level_counters(
		        PCL_state_descr *descr,
			int *counter_list,	/* I: events to be started */
			int nevents,   /* I: number of events */
			PCL_result_descr_t results[PCL_MAX_COUNTER], /* O: cntr descr */
			PCL_BIT_MASK_TYPE *counter_used_mask /* O: bit set to 1 if counter is used */
)

{
  int i, j, ecount, found;
  const PCL_event_descr_t *event;
  PCL_result_descr_t *this_result;
  unsigned int tmp_mask;


  for (j = 0; j < PCL_MAX_COUNTER; ++j)
    results[j].command = 4711;

  ecount = 0;

  /* generate bit mask what hardware counters are used */
  tmp_mask = 0;


  /*--------------------------------------------------------------------------*/
  /* handle each composed request */

  for (i = 0; i < nevents; ++i)
    {
      /* get event descriptor entry for i-th event request */
      event = &descr->event_table[counter_list[i]];

      /* check if event is supported */
      if ((event->available == PCL_AVAIL) && (event->composed == COMPOSED))
	{
	  /* check for low level counters which match event */
	  for (j = 0; j < PCL_MAX_COUNTER; ++j)
	    {
	      if (event->data.composed_data.command[j] != -1)
		{
		  if (!IS_USED(tmp_mask, j))
		    {
		      this_result = &results[j];

		      /* counter found */
		      /* this counter will be  used */
		      DEFINE_USE(tmp_mask, j);

		      /* command to activate counter */
		      this_result->command = event->data.composed_data.command[j];

		      /* counter value is used for result i */
		      this_result->result_index = i;

		      /* combine operation */
		      this_result->operation =
			event->data.composed_data.operation[j];

		      /* store to what event type it belongs */
		      this_result->event = counter_list[i];
		    }

		  else
		    /* counter already used */
		    return PCL_NOT_SUPPORTED;
		}		       /* if(event->data.composed... */
	    }			       /* for(j = 0; j < PCL_MAX_COUNTER; ++j) */

	  /* one more event done */
	  ++ecount;
	}
    }



  /*--------------------------------------------------------------------------*/
  /* handle each integral request                                             */
  /* this is done in two steps: first map events that can only run on one     */
  /* specific counter only. In the second step map the rest of the events     */
  /* might run on more than one counter.                                      */

  /* step 1 */
  for (i = 0; i < nevents; ++i)
    {
      /* get event descriptor entry for i-th event request */
      event = &descr->event_table[counter_list[i]];

      /* check if event is supported and not composed */
      if ((event->available == PCL_AVAIL) && (event->composed == NOT_COMP))
	{
	  /* check if the event must run on one specific counter (and this
	     counter is still available) */
	  found = 0;
	  for (j = 0; j < PCL_MAX_COUNTER; ++j)
	    {
	      if ((event->data.command[j] != -1) &&
		  (!IS_USED(tmp_mask, j)))
		++found;
	    }
	  if(found != 1)
	    continue;

	  /* check for low level counter which matches event and  which is
	     unused */
	  found = 0;
	  for (j = 0; (j < PCL_MAX_COUNTER) && !found; ++j)
	    {
	      if ((event->data.command[j] != -1) &&
		  (!IS_USED(tmp_mask, j)))
		{
		  /* this counter will be  used */
		  DEFINE_USE(tmp_mask, j);

		  /* counter found */
		  this_result = &results[j];

		  /* command to activate counter */
		  this_result->command = event->data.command[j];

		  /* counter value is used for result i */
		  this_result->result_index = i;

		  /* integral value (add value)  */
		  this_result->operation = OP_ADD;

		  /* store to what event type it belongs */
		  this_result->event = counter_list[i];

		  /* one more event done */
		  ++ecount;

		  /* leave loop */
		  found = 1;
		}
	    }
	}
    }


  /* step 2 */
  for (i = 0; i < nevents; ++i)
    {
      /* get event descriptor entry for i-th event request */
      event = &descr->event_table[counter_list[i]];

      /* check if event is supported and not composed */
      if ((event->available == PCL_AVAIL) && (event->composed == NOT_COMP))
	{
	  /* check if the event can run on more than one counter */
	  found = 0;
	  for (j = 0; j < PCL_MAX_COUNTER; ++j)
	    {
	      if ((event->data.command[j] != -1))
		++found;
	    }
	  if(found < 2)
	    continue;

	  /* check for low level counter which matches event and  which is
	     unused */
	  found = 0;
	  for (j = 0; (j < PCL_MAX_COUNTER) && !found; ++j)
	    {
	      if ((event->data.command[j] != -1) &&
		  (!IS_USED(tmp_mask, j)))
		{
		  /* this counter will be  used */
		  DEFINE_USE(tmp_mask, j);

		  /* counter found */
		  this_result = &results[j];

		  /* command to activate counter */
		  this_result->command = event->data.command[j];

		  /* counter value is used for result i */
		  this_result->result_index = i;

		  /* integral value (add value)  */
		  this_result->operation = OP_ADD;

		  /* store to what event type it belongs */
		  this_result->event = counter_list[i];

		  /* one more event done */
		  ++ecount;

		  /* leave loop */
		  found = 1;
		}
	    }
	}
    }


  /*--------------------------------------------------------------------------*/
  /* check if we got the whole mapping */

  if (ecount != nevents)
    /* at least one event is not supported */
    return PCL_NOT_SUPPORTED;
  
  /* store what counters are used */
  *counter_used_mask = tmp_mask;

  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* Based on the result values of the low-level counters, calculate results.   */


static void
generate_results(
		  PCL_state_descr *descr,
		  PCL_CNT_TYPE * results_list,             /* O: counter values */
		  int ncounter,                            /* I: number of counter values */
		  PCL_result_descr_t results[PCL_MAX_COUNTER]  /* I: result descr */
		  )

{
  int j;
  unsigned int op;
  PCL_result_descr_t *this_result;
  PCL_CNT_TYPE *reslist_ind, *ptr_value, value;
  double dvalue;


  /* initialize result array */
  for(j = 0; j < ncounter; j++)
    results_list[j] = 0;

  /* loop over all counters */
  for (j = 0; j < PCL_MAX_COUNTER; j++)
    {
      if(IS_USED(descr->counter_used_mask, j))
	{
	  this_result = &results[j];

	  /* get operation specifier */
	  op = this_result->operation;

	  /* get pointer to i-th counter value */
	  reslist_ind = &results_list[this_result->result_index];

	  /* get pointer to this result */
	  ptr_value = &(this_result->result_value);

	  if(op & OP_FP_TEST_MASK)
	    {
	      /* change 1st operand type from int to double? */
	      if (op & OP_FP_1)
		*((double *)reslist_ind) = (double)*reslist_ind;
	      
	      /* change 2nd operand type from int to double? */
	      if (op & OP_FP_2)
		*((double *)ptr_value) = (double)*ptr_value;
	      
	      /* mask out conversion operations */
	      op &= OP_FP_MASK;
	    }


	  /* get shortcuts */
	  value = *ptr_value;
	  dvalue = *((double *)ptr_value);

	  /* apply operation */
	  switch(op)
	    {
	    case OP_ADD:
	      /* add to result value the value of the j-th low-level counter */
	      *reslist_ind += value;
	      break;

	    case OP_SUB:
	      /* subtract from result value the value of the
		 j-th low-level counter
		 */
	      *reslist_ind -= value;
	      break;

	    case OP_FP_STORE:
	      /* store value in j-th low-level counter (floating point) */
	      *((double *)reslist_ind) = dvalue;
	      break;

	    case OP_FP_ADD:
	      /* add value in j-th low-level counter to value
		 (floating point) */
	      *((double *)reslist_ind) += dvalue;
	      break;

	    case OP_FP_MUL:
	      /* multiply value by the value of the j-th low-level
		 counter (floating point) */
	      *((double *)reslist_ind) *= dvalue;
	    break;

	    case OP_FP_DIV:
	      /* divide value by the value of the j-th low-level
		 counter (floating point) */
	      if(dvalue == 0.0)
		dvalue = 1.0;
	      *((double *)reslist_ind) /= dvalue;
	      break;

	    case OP_FP_DIV_MHZ:
	    {
	      /* divide value by value in the j-th low-level counter
		 and multiply by MHz-rate (floating point) */
	      if(dvalue <= 0.0)
		dvalue = 1.0;
	      *((double *)reslist_ind) *= descr->mhz_rate / dvalue;
	    }

	    case OP_FP_REZ:
	      /* add value by the repricoc value of the
		 j-th low-level counter (floating point) */
	      if(dvalue <= 0.0)
		dvalue = 1.0;
	      *((double *)reslist_ind) += 1.0 / dvalue;
	      break;

	    case OP_FP_REZ_MHZ:
	      /* add value by the repricoc value of the
		 j-th low-level counter and multiply by MHz-rate
		 (floating point) */
	      if(dvalue <= 0.0)
		dvalue = 1.0;
	      *((double *)reslist_ind) += descr->mhz_rate / dvalue;
	      break;

	    case OP_NOOP:
	      /* nothing to do */
	      break;

	    default:
	      fprintf(stderr, "PCL: internal error (illegal operation %d)\n", op);
	    }
	}
    }
}


/*                                                                            */
/*============================================================================*/
/* check input arguments to PCLquery and PCLstart                             */


static int
check_and_init(
	       PCL_state_descr *descr, /* I: handle */
	       int *counter_list,      /* I: event counters to be started */
	       int ncounter,	       /* I: number of event counters */
	       unsigned int flags      /* I: mode flags (PCL_MODE_xxx) */
	       )

{
#if defined(CHECK_ARGUMENTS)
  int i;
#endif


  /*--------------------------------------------------------------------------*/
  /* at this point the package must be initialized already */

  if (!descr->initialized)
    {
      int res;

      /* do this only once*/
      descr->initialized = 1;

      /* hardware dependend data initialization */
      if((res = PCL_driver_open(descr)) != PCL_SUCCESS)
	return res;

      /* determine MHz-rate of processor */
      descr->mhz_rate = PCL_determine_mhz_rate();
    }


  /*--------------------------------------------------------------------------*/
  /* check call arguments */

#if defined(CHECK_ARGUMENTS)

  /* check that number of requested events is smaller than maximum allowed
     number */
  if (ncounter >= PCL_MAX_EVENT_PER_CALL)
    return PCL_TOO_MANY_EVENTS;

  /* check that event identifiers are in allowed range */
  for (i = 0; i < ncounter; ++i)
    if ((counter_list[i] < 0) || (counter_list[i] >= PCL_MAX_EVENT))
      return PCL_ILL_EVENT;

  /* check correct mode */
  if((flags != PCL_MODE_USER)
     && (flags != PCL_MODE_SYSTEM)
     && (flags != PCL_MODE_USER_SYSTEM))
    return PCL_MODE_NOT_SUPPORTED;

  /* check if count mode is supported */
  if((flags & descr->allowed_count_mode) != flags)
    return PCL_MODE_NOT_SUPPORTED;
#endif


  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* conversion routines betwen a 64-bit int and a PCL_DESCR_TYPE               */


pcl_64int_t
map_descr_to_64int(PCL_DESCR_TYPE adr)
{
  if(sizeof(PCL_DESCR_TYPE) == sizeof(pcl_64int_t))
    return (pcl_64int_t)adr;
  else
    return (pcl_64int_t)(unsigned int)adr;
}

PCL_DESCR_TYPE
map_64int_to_descr(pcl_64int_t i64)
{
  if(sizeof(PCL_DESCR_TYPE) == sizeof(pcl_64int_t))
    return (PCL_DESCR_TYPE)i64;
  else
    return (PCL_DESCR_TYPE)(unsigned int)i64;
}


/*                                                                            */
/*============================================================================*/
/* initialize a thread-private state descriptor                               */


/* Fortran version */
int
fname(pclinit, PCLINIT)(pcl_64int_t *adr_descr)
{
  int res;
  PCL_DESCR_TYPE d;
 

  if((res = PCLinit(&d)) == PCL_SUCCESS)
    *adr_descr= map_descr_to_64int(d);

  return res;
}


/* Java version */
#if defined(JAVA_INTERFACE)
JNIEXPORT jint JNICALL
Java_PCL_cPCLinit(
		  JNIEnv *env,
		  jobject obj,
		  jlongArray adr_descr /* I/O: handle */
		  )
{
  jsize len;
  jlong *body;
  int res;
  PCL_DESCR_TYPE d;


  /* get array descriptor */
  len = (*env)->GetArrayLength(env, adr_descr);
  body = (*env)->GetLongArrayElements(env, adr_descr, 0);

  /* call PCL function */
  res = PCLinit(&d);
  *body = map_descr_to_64int(d);

  /* release copied array */
  (*env)->ReleaseLongArrayElements(env, adr_descr, body, 0);

  return res;
}
#endif


int
PCLinit(
	  PCL_DESCR_TYPE *adr_descr        /* I/O: address of handle */
	  )

{
  PCL_state_descr *d;
#if defined(PCL_MULTITHREADED)
  static int not_initialized = 1;
  static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;
#endif


  /* allocate thread-specific memory */
  d = calloc(1, sizeof(*d));
  if(d == NULL)
    return PCL_FAILURE;


  init_eventnames();

  /* initialize structure members */
  d->magic = MAGIC;
  d->virtual_counters_used_mask = 0;
  d->count_mode = PCL_MODE_USER;
  d->initialized = 0;
  d->nesting_level = -1;
  d->counter_used_mask = 0;
  d->fp_result_mask = 0;
  d->mhz_rate = -1.0;

  /* hardware dependend data initialization is done later in pcl_driver_open */
  *adr_descr = d;

#if defined(PCL_MULTITHREADED)
  /* lock mutex */
  pthread_mutex_lock(&lock);

  /* we are the first thread arriving? */
  if(not_initialized)
    {
      /* allocate thread specific key */
      pthread_key_create(&descr_key, NULL);

      not_initialized = 0;
    } 

  /* unlock mutex */
  pthread_mutex_unlock(&lock);

  /* store the descriptor as thread-specific data */
  pthread_setspecific(descr_key, d);

#else
  /* store the descriptor */
  state_descr = d;
#endif


  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* release thread-specific descriptor                                         */


/* Fortran version */
int
fname(pclexit, PCLEXIT)(pcl_64int_t *descr)
    
{
  return PCLexit(map_64int_to_descr(*descr));
}


/* Java version */
#if defined(JAVA_INTERFACE)
JNIEXPORT jint JNICALL
Java_PCL_cPCLexit(
		  JNIEnv *env,
		  jobject obj,
		  jlong descr            /* I: handle */
		  )
{
  /* call PCL function */
  return PCLexit(map_64int_to_descr(descr));
}
#endif


int
PCLexit(
	PCL_DESCR_TYPE descr        /* I: handle */
	)

{
  PCL_state_descr *d;

  d = descr;

  /* minor check for consistency */
  if(d->magic != MAGIC)
    {
      free(d);
      return PCL_FAILURE;
    }
  else
    {
      free(d);
      return PCL_SUCCESS;
    }
}


/*                                                                            */
/*============================================================================*/
/* query for functionality.                                                   */
/* returns PCL_SUCCESS if requested event counting is possible                */


/* Fortran version */
int
fname(pclquery, PCLQUERY)
     (
      pcl_64int_t *descr,         /* I: handle */
      int *counter_list,	  /* I: requested event counters */
      int *ncounter,	          /* I: # of requested event counters */
      unsigned int *flags         /* I: mode flags (PCL_MODE_xxx) */
      )

{
  return PCLquery(map_64int_to_descr(*descr), counter_list, *ncounter, *flags);
}


/* Java version */
#if defined(JAVA_INTERFACE)
JNIEXPORT jint JNICALL
Java_PCL_cPCLquery(
		  JNIEnv *env,
		  jobject obj,
		  jlong descr,            /* I: handle */
		  jintArray counter_list, /* I: requested event counters */
		  jint ncounter,          /* I: # of requested event counters */
		  jint flags              /* I: mode flags (PCL_MODE_xxx) */
		  )

{
  jsize len;
  jint *body;
  int res;

  /* get array descriptor */
  len = (*env)->GetArrayLength(env, counter_list);
  body = (*env)->GetIntArrayElements(env, counter_list, 0);

  /* call PCL function */
  res = PCLquery(map_64int_to_descr(descr), (int *)body, ncounter, flags);

  /* release copied array */
  (*env)->ReleaseIntArrayElements(env, counter_list, body, 0);

  return res;
}
#endif


/* C, C++ version */
int
PCLquery(
	   PCL_DESCR_TYPE descr,       /* I: handle */
	   int *counter_list,	       /* I: requested event counters */
	   int ncounter,	       /* I: # of requested event counters */
	   unsigned int flags          /* I: mode flags (PCL_MODE_xxx) */
)

{
  int res;
  PCL_result_descr_t tmp_results[PCL_MAX_COUNTER];
  PCL_BIT_MASK_TYPE tmp_mask;
  PCL_state_descr *d;
	     

  d = (PCL_state_descr *) descr;

  /* check call arguments */
  if((res = check_and_init(d, counter_list, ncounter, flags)) != PCL_SUCCESS)
    return res;

  /* Try to get a mapping of events to counters in a temporary counter list.
     If this can be done, the query is successful.
  */
  return map_low_level_counters(d, counter_list, ncounter, tmp_results,
				&tmp_mask);
}


/*                                                                            */
/*============================================================================*/
/* start performance counting                                                 */
/* returns PCL_SUCCESS if everything went ok                                  */
/* If this is a nested call we do the following:                              */
/*    - increment the nesting level                                           */
/*    - check that same events are requested than in previous level           */
/*    - stops previous couting and store result values in <partial_results>   */


/* Fortran version */
int
fname(pclstart, PCLSTART)(
           pcl_64int_t *descr,         /* I: handle */
	   int *counter_list,	       /* I: event counters to be started */
	   int *ncounter,	       /* I: number of event counters */
	   unsigned int *flags         /* I: mode flags (PCL_MODE_xxx) */
)

{
  return PCLstart(map_64int_to_descr(*descr), counter_list, *ncounter, *flags);
}


/* Java version */
#if defined(JAVA_INTERFACE)
JNIEXPORT jint JNICALL
Java_PCL_cPCLstart(
		  JNIEnv *env,
		  jobject obj,
		  jlong descr,            /* I: handle */
		  jintArray counter_list, /* I: requested event counters */
		  jint ncounter,          /* I: # of requested event counters */
		  jint flags              /* I: mode flags (PCL_MODE_xxx) */
		  )

{
  jsize len;
  jint *body;
  int res;


  /* get array descriptor */
  len = (*env)->GetArrayLength(env, counter_list);
  body = (*env)->GetIntArrayElements(env, counter_list, 0);

  /* call PCL function */
  res = PCLstart(map_64int_to_descr(descr), (int *)body, ncounter, flags);

  /* release copied array */
  (*env)->ReleaseIntArrayElements(env, counter_list, body, 0);

  return res;
}
#endif


/* C, C++ version */
int
PCLstart(
	 PCL_DESCR_TYPE descr,       /* I: handle */
	 int *counter_list,	     /* I: event counters to be started */
	 int ncounter,	             /* I: number of event counters */
	 unsigned int flags          /* I: mode flags (PCL_MODE_xxx) */
)

{
  int i, res, nesting_level;
  PCL_state_descr *d = descr;


  /*--------------------------------------------------------------------------*/
  /* check call arguments */

  if((res = check_and_init(d, counter_list, ncounter, flags)) != PCL_SUCCESS)
    return res;

#if defined(CHECK_ARGUMENTS)
  /* check for incrementing nesting */
  if (d->nesting_level + 1 >= PCL_MAX_NESTING_LEVEL)
    return PCL_TOO_MANY_NESTINGS;
#endif


  /*--------------------------------------------------------------------------*/
  /* initialize nesting level */

  /* increment nesting level */
  nesting_level = ++d->nesting_level;

  /* so far, the virtual counters for this nesting level are not used */
  RESET_USE(d->virtual_counters_used_mask, nesting_level);

  /* set virtual counters and start values for this nesting level to 0 */
  for(i = 0; i < PCL_MAX_COUNTER; ++i)
    {
      d->virtual_counters[nesting_level][i] = 0;
      d->start_values[nesting_level][i] = 0;
    }


  /*--------------------------------------------------------------------------*/
  /* handle previous calls */

  if (nesting_level == 0)
    {
      /* outermost nesting level */
      /* store request for this call */
      d->event_count = ncounter;
      for (i = 0; i < ncounter; ++i)
	d->events[i] = counter_list[i];

      /* store count mode */
      d->count_mode = flags;

      /* try to get a mapping of events to counters */
      if((res = map_low_level_counters(d, counter_list, ncounter,
				       d->results,
				       &d->counter_used_mask))
	 != PCL_SUCCESS)
	return res;

      d->fp_result_mask = 0;
      for(i = 0; i < ncounter; ++i)
	if(counter_list[i] >= PCL_MFLOPS)
	  DEFINE_USE(d->fp_result_mask, i);

      /* start the counters */
      if((res = hw_counter_start(d)) != PCL_SUCCESS)
	return res;

      /* read out current values as base values for this nesting level */
      if((res = hw_counter_read(d, d->start_values[nesting_level]))
	 != PCL_SUCCESS)
	return res;
    }

  else
    {
      /* inner nesting level */

      /* check that the same events are requested as in lower nesting levels */
#if defined(CHECK_ARGUMENTS)
      /* check number of requests */
      if (ncounter != d->event_count)
	return PCL_ILL_NESTING;

      /* check that event requests of this call are the same than that of
         previous call */
      for (i = 0; i < ncounter; ++i)
	if (counter_list[i] != d->events[i])
	  return PCL_ILL_NESTING;

      /* check correct mode */
      if(flags != d->count_mode)
	return PCL_MODE_NOT_SUPPORTED;
#endif

      /* read out current values as base values for this nesting level */
      if((res = hw_counter_read(d, d->start_values[nesting_level]))
	 != PCL_SUCCESS)
	return res;
    }


  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* read out counter values and process values                                 */


static int
do_read_or_stop(
		PCL_state_descr *descr,
		PCL_CNT_TYPE * i_result_list,     /* O: integer counter values */
		PCL_FP_CNT_TYPE * fp_result_list, /* O: FP counter values */
		int ncounter,                     /* I: number of event counters */
		int do_restart                    /* I: 1=restart counting, 0=stop */
		)


{
  int i, res, nesting_level;
  PCL_CNT_TYPE counter_values[PCL_MAX_COUNTER];


  /*--------------------------------------------------------------------------*/
  /* check call arguments */

  nesting_level = descr->nesting_level;
#if defined(CHECK_ARGUMENTS)
  /* check call level and requests for this call with corresponding PCLstart */
  if ((nesting_level < 0) || (descr->event_count != ncounter))
    return PCL_ILL_NESTING;
#endif


  /*--------------------------------------------------------------------------*/
  /* read values */

  if(do_restart)
    {
      /* read out values */
      if ((res = hw_counter_read(descr, counter_values)) != PCL_SUCCESS)
	return res;
    }
  else
    {
      /* read out values and stop counters */
      if ((res = hw_counter_read_and_stop(descr, counter_values) != PCL_SUCCESS))
	return res;
    }


  /* subtract start values gotten with PCLstart and add partial results */
  if(IS_USED(descr->virtual_counters_used_mask, nesting_level))
    {
      for (i = 0; i < PCL_MAX_COUNTER; ++i)
	if(IS_USED(descr->counter_used_mask, i))
	  descr->results[i].result_value =
	    (counter_values[i] - descr->start_values[nesting_level][i])
	    + descr->virtual_counters[nesting_level][i];
    }
  else
    {
      for (i = 0; i < PCL_MAX_COUNTER; ++i)
	if(IS_USED(descr->counter_used_mask, i))
	  descr->results[i].result_value =
	    counter_values[i] - descr->start_values[nesting_level][i];
    }

  /* generate results from performance counter values */
  generate_results(descr, i_result_list, ncounter, descr->results);


  /* transfer floating point results to fp_result_list */
  if(descr->fp_result_mask)
    {
      for(i = 0; i < descr->event_count; ++i)
	if(IS_USED(descr->fp_result_mask, i))
	  fp_result_list[i] = *(PCL_FP_CNT_TYPE *)&i_result_list[i];
    }


  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* reads performance counters                                                 */
/* result_list must be a pointer to a sufficiently large buffer.              */
/* returns PCL_SUCCESS if everything went ok                                  */


/* Fortran version */
int
fname(pclread, PCLREAD)(
	  pcl_64int_t *descr,               /* I: handle */
	  PCL_CNT_TYPE * i_result_list,     /* O: integer counter values */
	  PCL_FP_CNT_TYPE * fp_result_list, /* O: FP counter values */
	  int *ncounter		            /* I: number of event counters */
)

{
  return PCLread(map_64int_to_descr(*descr), i_result_list, fp_result_list, *ncounter);
}


/* Java version */
#if defined(JAVA_INTERFACE)
JNIEXPORT jint JNICALL
Java_PCL_cPCLread(
		  JNIEnv *env,
		  jobject obj,
		  jlong descr,                 /* I: handle */
		  jlongArray i_result_list,    /* O: integer counter values */
		  jdoubleArray fp_result_list, /* O: FP counter values */
		  jint ncounter                /* I: number of event counters */
		  )

{
  jsize i_len, fp_len;
  jlong *i_body;
  jdouble *fp_body;
  int res;


  /* get array descriptor */
  i_len = (*env)->GetArrayLength(env, i_result_list);
  fp_len = (*env)->GetArrayLength(env, fp_result_list);
  i_body = (*env)->GetLongArrayElements(env, i_result_list, 0);
  fp_body = (*env)->GetDoubleArrayElements(env, fp_result_list, 0);

  /* call PCL function */
  res = PCLread(map_64int_to_descr(descr), i_body, fp_body, ncounter);

  /* release copied array */
  (*env)->ReleaseLongArrayElements(env, i_result_list, i_body, 0);
  (*env)->ReleaseDoubleArrayElements(env, fp_result_list, fp_body, 0);

  return res;
}
#endif


/* C, C++ version */
int
PCLread(
	PCL_DESCR_TYPE descr,             /* I: handle */
	PCL_CNT_TYPE * i_result_list,     /* O: integer counter values */
	PCL_FP_CNT_TYPE * fp_result_list, /* O: FP counter values */
	int ncounter		          /* I: number of event counters */
)

{
  /* get results */
  return do_read_or_stop(descr, i_result_list, fp_result_list, ncounter, 1);
}


/*                                                                            */
/*============================================================================*/
/* stops performance counting and reads counters values.                      */
/* result_list must be a pointer to a sufficiently large buffer.              */
/* returns PCL_SUCCESS if everything went ok                                  */


/* Fortran version */
int
fname(pclstop, PCLSTOP)(
          pcl_64int_t *descr,                /* I: handle */
	  PCL_CNT_TYPE * i_result_list,      /* O: integer counter values */
	  PCL_FP_CNT_TYPE * fp_result_list,  /* O: FP counter values */
	  int *ncounter		             /* I: number of event counters */
)

{
  return PCLstop(map_64int_to_descr(*descr), i_result_list, fp_result_list, *ncounter);
}


/* Java version */
#if defined(JAVA_INTERFACE)
JNIEXPORT jint JNICALL
Java_PCL_cPCLstop(
		  JNIEnv *env,
		  jobject obj,
		  jlong descr,                  /* I: handle */
		  jlongArray i_result_list,     /* O: integer counter values */
		  jdoubleArray fp_result_list,  /* O: FP counter values */
		  jint ncounter                 /* I: number of event counters */
		  )

{
  jsize i_len, fp_len;
  jlong *i_body;
  jdouble *fp_body;
  int res;


  /* get array descriptor */
  i_len = (*env)->GetArrayLength(env, i_result_list);
  fp_len = (*env)->GetArrayLength(env, fp_result_list);
  i_body = (*env)->GetLongArrayElements(env, i_result_list, 0);
  fp_body = (*env)->GetDoubleArrayElements(env, fp_result_list, 0);

  /* call PCL function */
  res = PCLstop(map_64int_to_descr(descr), i_body, fp_body, ncounter);

  /* release copied array */
  (*env)->ReleaseLongArrayElements(env, i_result_list, i_body, 0);
  (*env)->ReleaseDoubleArrayElements(env, fp_result_list, fp_body, 0);
  return res;
}
#endif


/* C, C++ version */
int
PCLstop(
	PCL_DESCR_TYPE descr,              /* I: handle */
	PCL_CNT_TYPE * i_result_list,      /* O: integer counter values */
	PCL_FP_CNT_TYPE * fp_result_list,  /* O: FP counter values */
	int ncounter		           /* I: number of event counters */
)

{
  int i, res, nesting_level;
  PCL_state_descr *d = descr;
#if defined(COUNTER_SIZE_32) && defined(COUNTER32_ARE_NOT_ENOUGH)
  /* 32-bit performance counters */
  struct itimerval itimer;
#endif


  /*--------------------------------------------------------------------------*/

  /* check arguments and read out values */
  nesting_level = d->nesting_level;
  if((res = do_read_or_stop(d, i_result_list, fp_result_list, ncounter,
			    nesting_level > 0)) != PCL_SUCCESS)
    return res;

  if (nesting_level > 0)
    {
      /* add virtual counter values to previous nesting level */
      if(IS_USED(d->virtual_counters_used_mask, nesting_level))
	for(i = 0; i < PCL_MAX_COUNTER; ++i)
	  d->virtual_counters[nesting_level-1][i] += d->virtual_counters[nesting_level][i];
    }

  /* decrement nesting level*/
  nesting_level = --d->nesting_level;


  /*--------------------------------------------------------------------------*/
  /* stop timer interrupts (overflow avoidance) */

#if defined(COUNTER_SIZE_32) && defined(COUNTER32_ARE_NOT_ENOUGH)
  /* 32-bit performance counters */
  if (nesting_level == -1)
    {
      /* set up signal handler */
      if (signal(ITIMER_SIGNAL, d->old_signal_handler) == SIG_ERR)
	return PCL_FAILURE;

      /* cancel interval timer */
      itimer.it_interval.tv_sec = 0;
      itimer.it_interval.tv_usec = 0;
      itimer.it_value.tv_sec = 0;
      itimer.it_value.tv_usec = 0;
      setitimer(ITIMER_MODE, &itimer, NULL);

      d->old_signal_handler = NULL;
    }
#endif


  /*--------------------------------------------------------------------------*/
  /* leave funtion */

  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* returns name (character string) of event.           .                      */


/* Fortran version */
const char *
fname(pcleventname, PCLEVENTNAME)(
		  int *event		       /* I: event to get name for */
                 )

{
    /* ??? not yet done
     Fortran character passing is one of the few unsolved problems
     in computer science. We don't solve it for the moment.
    */
    return PCLeventname(*event);
}


/* Java version */
#if defined(JAVA_INTERFACE)
JNIEXPORT jstring JNICALL
Java_PCL_cPCLeventname(
		       JNIEnv *env,
		       jobject obj,
		       jint event           /* I: event to get name for */
		       )

{
  const char *res;

  /* call PCL function */
  res = PCLeventname(event);
  if(res == NULL)
    /* should not happen */
    res = "<unkown>";

  return (*env)->NewStringUTF(env, res);
}
#endif


const char *
PCLeventname(
	     int event		       /* I: event to get name for */
)

{
  if ((event < 0) || (event >= PCL_MAX_EVENT))
    return (const char *)NULL;
  else
    return PCL_eventname[event];
}


/*                                                                            */
/*============================================================================*/
/* Return the value of an elapsed cycle counter if available, otherwise -1.   */
/* The functionality exists also with PCLstart/PCLstop but sometimes          */
/* it is important to have a very fast function.                              */
/* Currently, this is not part of the official PCL interface.                 */

#if 0

PCL_CNT_TYPE
PCLelapsed_cycle_count(void)
{
#if defined(ALPHA_CRAY)
  return ev5_read_cycle_counter();

#elif defined(ALPHA_DEC)
  return read_cycle_counter();

#elif defined(SPARC)
  return get_tick();

#elif defined(MIPS)
  /*???*/
  return -1;

#elif defined(POWERPC)
  /*???*/
  return -1;

#elif defined(PENTIUM_LINUX)
  {
  long long clock_value;
  
  /* fast assembler code to access the cycle counter (TSC) */
  asm volatile ("rdtsc" : "=a"(*(unsigned int *)&clock_value),
		          "=d"(*(((unsigned int *)&clock_value)+1)));

  return clock_value;
  }
#endif
}
#endif


/*============================================================================*
 *                             that's all folks                               *
 *============================================================================*/

