/*==============================================================================

   Purpose:    small test program for PCL's low-level interface


   Author:     Rudolf Berrendorf
               University of Applied Sciences Bonn-Rhein-Sieg
               Computer Science Department
               53754 Sankt Augustin
               rudolf.berrendorf@fh-bonn-rhein-sieg.de


   last mod:   19.10.2002



==============================================================================*/


#include "../src/pcl.h"
#include "../src/determine_mhz.h"

/*============================================================================*/

#define N 200

/*                                                                            */
/*============================================================================*/
/* matrix addition as a simple test program.                                  */

static void
matinit(double a[N][N], double b[N][N], double c[N][N])
{
  int i,j;

  for (i = 0; i < N; ++i)
    for (j = 0; j < N; ++j)
      a[i][j] = b[i][j] = c[i][j] = 1.0;
}

static void
matadd(double a[N][N], double b[N][N], double c[N][N])

{
  int i, j;

  for (i = 0; i < N; ++i)
    for (j = 0; j < N; ++j)
      a[i][j] = b[i][j] + c[i][j];
}


/*============================================================================*/

int
main(int argc, char **argv)

{
  int res, counter_index;
  unsigned int count_mode;
  double a[N][N], b[N][N], c[N][N];
  char *output;
  PCL_BIT_MASK_TYPE counter_used_mask;
  PCL_DRIVER_COMMAND_TYPE commands[PCL_COUNTER_MAX];
  PCL_CNT_TYPE counter_values[PCL_COUNTER_MAX];
  PCL_DESCR_TYPE descr;


  printf("\nProgram to test the low level interface of PCL.\n\n");

  /* default result output string */
  output = "%.0f floating point instructions, should be approx. %d (dummy output=%2.0f)\n";

  counter_used_mask = 0;
  /* initialize test arrays and do some initial work */
  matinit(a,b,c);
  matadd(a, b, c);


  /* allocate descriptor */
  if((res = PCLinit(&descr)) != PCL_SUCCESS)
    {
      printf("error on PCLinit (%d)\n", res);
      exit(1);
    }

  /* open driver */
  if((res = PCL_driver_open(descr)) != PCL_SUCCESS)
    {
      printf("error on PCL_driver_open (%d)\n", res);
      exit(1);
    }


  /* start counter: we count FP instructions */
#if defined(ALPHA_CRAY)
  counter_index = 1;
  commands[counter_index] = 0x0a;

#elif defined(ALPHA_DEC)
  counter_index = 1;
  commands[counter_index] = 0x0a;

#elif defined(MIPS)
  counter_index = 1;
  commands[counter_index] = 16+5;

#elif defined(SPARC)
  /* not FP instructions but rather instructions */
  counter_index = 0;
#if defined(PERFMON)
  commands[counter_index] = (0x1 << 4);
#else
  commands[counter_index] = 0x1;
#endif
  output = "%.0f instructions (dummy output=%d, %2.0f)\n";

#elif defined(POWERPC)
  /* this works only on Power3 */
  /* other processors need other definitions here */
  counter_index = 1;
  commands[counter_index] = 10;

#elif defined(PENTIUM_LINUX)
  /* for Pentium II and III */
  counter_index = 0;
  commands[counter_index] = 0xc1;
  /* for Pentium 4 */
  counter_index = 8;
  commands[counter_index] = (4 << 19) | (0x8000 << 3) | (1);

#elif defined(IA64_LINUX)
  /* for Intel IA64 */
  counter_index = 0;
  commands[counter_index] = 0x0a;

#elif defined(SR8000)
  counter_index = 7;
  commands[counter_index] = 1;

#else
#error "unsupported system"
#endif


  /* one counter used */
  counter_used_mask = (0x1 << counter_index);
  count_mode = PCL_MODE_USER;

  /* start driver */
  if((res = PCL_driver_start(descr, PCL_COUNTER_MAX, counter_used_mask,
			     commands, count_mode)) != PCL_SUCCESS)
    {
      printf("error on PCL_driver_start (%d)\n", res);
      exit(1);
    }

  /* do some work */
  matadd(a,b,c);

  /* read counter */
  res = PCL_driver_read(descr, PCL_COUNTER_MAX, counter_used_mask,
			    counter_values);

#if defined(PENTIUM_LINUX)
/* fast assembler code to access the cycle counter (TSC) */
#define rdtsc(valptr) asm volatile ("rdtsc" : \
"=a"(*(unsigned int *)valptr), "=d"(*(((unsigned int *)valptr)+1)))
 {
   /* measure overhead */
   long long t0, t1;
   int mhz;

   if((mhz = PCL_determine_mhz_rate()) == -1)
     mhz = 1;
   rdtsc(&t0);
   res = PCL_driver_read(descr, PCL_COUNTER_MAX, counter_used_mask,
			 counter_values);
   rdtsc(&t1);
   printf("overhead reading counters: %d cycles (%.0f ns)\n\n",
	  (int)(t1-t0), (t1-t0) / (double)mhz * 1e3);
 }
#endif

  if(res != PCL_SUCCESS)
    {
      printf("error on PCL_driver_read (%d)\n", res);
      exit(1);
    }
  else
    printf(output, (double)counter_values[counter_index], N*N, a[0][0]);

  /* stop counting */
  if((res = PCL_driver_stop(descr)) != PCL_SUCCESS)
    {
      printf("error on PCL_driver_stop (%d)\n", res);
      exit(1);
    }

  /* close driver */
  if((res = PCL_driver_close(descr)) != PCL_SUCCESS)
    {
      printf("error on PCL_driver_close (%d)\n", res);
      exit(1);
    }

  /* release descriptor */
  if((res = PCLexit(descr)) != PCL_SUCCESS)
    {
      printf("error on PCLexit_r (%d)\n", res);
      exit(1);
    }

  return 0;
}


/*============================================================================*
 *                             that's all folks                               *
 *============================================================================*/
