/*==============================================================================


   Purpose:    test program for Performance Counter Library (PCL)
               tesing the thread-safe interface


   Author:     Rudolf Berrendorf
               University of Applied Sciences Bonn-Rhein-Sieg
               Computer Science Department
               53754 Sankt Augustin
               rudolf.berrendorf@fh-bonn-rhein-sieg.de


   last mod:   20.10.2002


==============================================================================*/

#include <pthread.h>
#include <errno.h>
#include <string.h>

#include "../src/pcl.h"

#define NTHREADS 4          /* number of threads */
#define N        200        /* matrix dimension */

static double a[N][N], b[N][N], c[N][N];
static PCL_PROCESSOR_INFO info;

/*----------------------------------------------------------------------------*/
/* some work */

void do_work()
{
  int i,j;

  for(i=0; i<N; i++)
    for(j=0; j<N; j++)
      a[i][j] = b[i][j] + c[i][j];
}

/*----------------------------------------------------------------------------*/

void *count(void *arg)
{
  int threadnum, i;
  int counter_list[2];
  int ncounter, res;
  unsigned int mode;
  PCL_CNT_TYPE i_result_list[2];
  PCL_FP_CNT_TYPE fp_result_list[2];
  PCL_DESCR_TYPE descr;


  threadnum = *(int *)arg;

  /* get descriptor */
  if(PCLinit(&descr) != PCL_SUCCESS)
    {
      printf("init not possible\n");
      return NULL;
    }

  /* Define what we want to measure (2 different things). */
  ncounter = 2;
  if(!strcmp(info.family, "Pentium 4"))
    counter_list[0] = PCL_ELAPSED_CYCLES;
  else
    counter_list[0] = PCL_CYCLES;
  if((threadnum % 2) == 0)
    counter_list[1] = PCL_INSTR;
  else
    counter_list[1] = PCL_FP_INSTR;

  /* define count mode */
  mode = PCL_MODE_USER;
  
  /* Check if this is possible on the machine. */
  if((res = PCLquery(descr, counter_list, ncounter, mode)) != PCL_SUCCESS)
    {
      printf("requested events not possible (retval=%d, threadnum=%d)\n",
	     res, threadnum);
      return NULL;
    }

  /* Start performance counting.
     We have checked already the requested functionality
     with PCL_query, so no error check would be necessary. */
  if((res = PCLstart(descr, counter_list, ncounter, mode)) != PCL_SUCCESS)
    {
      printf("something went wrong (retval=%d, thread=%d)\n", res, threadnum);
      return NULL;
    }

  /* Here comes the work to be measured. */
  for(i = 0; i < N; i++)
    do_work();
  
  /* Stop performance counting and get the counter values. */
  if((res = PCLstop(descr, i_result_list, fp_result_list, ncounter))
     != PCL_SUCCESS)
    {
      printf("problems with stopping counters (retval=%d, threadnum=%d)\n",
	     res, threadnum);
      return NULL;
    }

  /* print out results */
  printf("thread-test(%d): %.0f %s in %.0f cycles",
	 threadnum,
	 (double)i_result_list[1],
	 ((threadnum % 2) == 0) ? "instructions" : "fp-instructions",
	 (double)i_result_list[0]);
  if((threadnum % 2) != 0)
    printf(" (FP instructions should be roughly %d)\n", N*N*N);
  else
    printf("\n");

  /* release descriptor */
  if((res = PCLexit(descr)) != PCL_SUCCESS)
    printf("exit  not possible (retval=%d, threadnum=%d)\n",
	   res, threadnum);

  return NULL;
}

/*----------------------------------------------------------------------------*/

int main(int argc, char **argv)
{
  int i;
  pthread_t thr[NTHREADS];
  pthread_attr_t attr[NTHREADS];
  int arg[NTHREADS];


  printf("\nProgram to test PCL with threads.\n"
	 "Threads 0 and 2 count cycles and instructions,\n"
	 "threads 1 and 3 count cycles and floating point instructions.\n"
	 "(Did you enable multithreading support in the configuration step?)\n");

  /* get processor information */
  if(PCL_driver_info(&info) != PCL_SUCCESS)
    printf("unable to get processor information\n");

  for(i = 0; i < NTHREADS; i++)
    {
      arg[i] = i;

      /* init attribute */
      if(pthread_attr_init(&attr[i]) != 0)
	printf("error in attr init for pthread\n");
	
      /* bind thread */
      if(pthread_attr_setscope(&attr[i], PTHREAD_SCOPE_SYSTEM) != 0)
	printf("error in setscope for pthread\n");

      if(pthread_create(&thr[i], &attr[i], count, &arg[i]) != 0)
      {
	perror("error creating pthread");
	exit(1);
      }
    }

  for(i = 0; i < NTHREADS; i++)
    pthread_join(thr[i], NULL);

  return 0;
}


/*============================================================================*
 *                             that's all folks                               *
 *============================================================================*/
